from glyles.glycans.factory.factory_f import FuranoseFactory
from glyles.glycans.factory.factory_o import c1_finder, OpenFactory
from glyles.glycans.factory.factory_p import PyranoseFactory
from glyles.glycans.monomer import Monomer
from glyles.glycans.utils import Config, Enantiomer, Lactole
from glyles.grammar.GlycanLexer import GlycanLexer


class MonomerFactory:
    """
    Class holding and managing the access and generation of all monomers implemented in this package.
    """

    def __init__(self):
        """
        Initialize this factory by creating instances of all "sub-"factories
        """
        self.pyranose_fac = PyranoseFactory()
        self.furanose_fac = FuranoseFactory()
        self.open_fac = OpenFactory()

        self.keys = set(self.pyranose_fac.keys()).union(self.furanose_fac.keys())

    def __contains__(self, item):
        """
        Check if an item is part of this factory and can be returned.

        Args:
            item (str): Code of a monomer to be checked

        Returns:
            True if the item is included in the current version of this package
        """
        return item.upper() in self.keys

    def __getitem__(self, item):
        """
        Get an instance of a monomer from this factory.

        Args:
            item (str): name of the query monomer

        Returns:
            Directory containing all necessary information to initialize a monomer implementation
        """
        furanose = item in ["ERY", "THRE", "RUL", "XUL", "XLU", "ACE"]
        if item[-1] == "p" and not item.endswith("Hep"):
            item = item[:-1]
        if item[-1] == "f":
            item = item[:-1]
            furanose = True

        if furanose and item in self.furanose_fac:
            return self.furanose_fac[item]
        if not furanose and item in self.pyranose_fac:
            return self.pyranose_fac[item]
        if item in self.open_fac:
            return self.open_fac[item]
        raise NotImplementedError("Query-monomer is neither in pyranoses nor in furanoses")

    def keys(self):
        """
        Get all monomers that are included in this package by their extended name.

        Returns:
            Set of names for all monomers, their available derivatives and configurations (alpha/beta/undefined)
        """
        return self.keys

    def monomers(self):
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations.

        Returns:
            List, sorted from long to short, of all monomer names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self.keys]), key=lambda x: -len(x))

    def furanoses(self):
        """
        Get the names of all monomers available as furanoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all furanose names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self.furanose_fac.keys()]))

    def pyranoses(self):
        """
        Get the names of all monomers available as pyranoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all pyranose names in upper case
        """
        return sorted(set([self[x.split("_")[-1]]["name"] for x in self.pyranose_fac.keys()]))

    def monomer_names(self):
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.monomers():
            if item in self.pyranose_fac:
                output.add(self.pyranose_fac[item]["name"])
        return list(output)

    def pyranose_names(self):
        """
        Get the names of all pyranoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.pyranoses():
            if item in self.pyranose_fac:
                output.add(self.pyranose_fac[item]["name"])
        return list(output)

    def furanose_names(self):
        """
        Get the names of all furanoses in this package ignoring the alpha/beta conformations

        Returns:
            List, sorted from long to short, of all monomer names in IUPAC notation
        """
        output = set()
        for item in self.monomers():
            if item in self.furanose_fac:
                output.add(self.furanose_fac[item]["name"])
        return list(output)

    @staticmethod
    def unknown_monomer(name):
        """
        Generate a dummy monosaccharide containing no information except the name as the structure and all its
        properties are unknown.
        Args:
            name (str): Name of the monosaccharide with structure and properties are unknown

        Returns:
            Dummy information about an unknown monomer
        """
        return {"name": name, "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.UNKNOWN,
                "smiles": ""}

    @staticmethod
    def succinic_acid():
        """
        This method stores the information about succinic acid in a dictionary.

        Returns:
            Information about succinic acid
        """
        return {"name": "Suc", "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.OPEN,
                "smiles": "OC[C@H](O)CCO", "c1_find": lambda x: c1_finder(x, "OC[C@H](O)CCO")}

    def create(self, recipe, config=None, tree_only=False):
        """
        Create a monomer from its describing IUPAC string with all added side chains.

        Args:
            recipe (List[Tuple[str, int]]): List of modifications, conformations and the root monomer
            config (str): configuration if monomer is alpha or beta monomer
            tree_only (bool): Flag indicating to only parse the tree of glycans and not the modifications

        Returns:
            Monomer-Instance containing all modifications given in the input
        """

        # extract key information from the input, i.e. the type, the configuration and pyranose/furanose
        tmp = list(zip(*recipe))
        name = recipe[tmp[1].index(GlycanLexer.SAC)][0]
        if name == "Sug":
            name = "Oct"
        config_index = tmp[1].index(GlycanLexer.TYPE) if GlycanLexer.TYPE in tmp[1] else None
        ring_index = tmp[1].index(GlycanLexer.RING) if GlycanLexer.RING in tmp[1] else None

        # generate the full name that is looked up in the factory
        if config is not None and len(config) > 0:
            name = config + "_" + name
        elif config_index is not None:
            name = recipe[config_index][0] + "_" + name

        # get the monomer from the factory
        if name in self.pyranose_fac and (ring_index is None or recipe[ring_index][0] != "f"):
            monomer = Monomer(**self.pyranose_fac[name], recipe=recipe)
        elif name in self.furanose_fac:
            monomer = Monomer(**self.furanose_fac[name], recipe=recipe)
        elif name in self.open_fac:
            monomer = Monomer(**self.open_fac[name], recipe=recipe)
        elif name[-3:].upper() == "SUC":
            monomer = Monomer(**self.succinic_acid(), recipe=recipe)
        else:
            monomer = Monomer(**self.unknown_monomer(name), recipe=recipe)

        full = False
        if not tree_only:
            # create the final molecule using the molecule's react-method augmented with the recipe of the molecule
            monomer, full = monomer.react(*tmp)

        # set full to false in case the monomer is unknown, i.e. neither pyranose nor furanose nor open-form
        full &= (monomer.get_lactole() != Lactole.UNKNOWN)

        return monomer, full



from glyles.glycans.utils import *


class FuranoseFactory:
    """
    Checked with http://www.cheminfo.org/flavor/malaria/Utilities/SMILES_generator___checker/index.html
    """

    __monomers = {
        # Heptose
        "HEP": {"name": "Hep", "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                "smiles": "OCC(O)C(O)C1OC(O)C(O)C1O"},
        "A_HEP": {"name": "Hep", "config": Config.ALPHA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C(O)C1O[C@H](O)C(O)C1O"},
        "B_HEP": {"name": "Hep", "config": Config.BETA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C(O)C1O[C@@H](O)C(O)C1O"},

        # Hexose
        "HEX": {"name": "Hex", "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                "smiles": "C(C(C1C(C(C(O1)O)O)O)O)O"},
        "A_HEX": {"name": "Hex", "config": Config.ALPHA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C1O[C@H](O)C(O)C1O"},
        "B_HEX": {"name": "Hex", "config": Config.BETA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C1O[C@@H](O)C(O)C1O"},

        # Octose
        "OCT": {"name": "Oct", "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                "smiles": "OCC(O)C(O)C(O)C1OC(O)C(O)C1O"},
        "A_OCT": {"name": "Oct", "config": Config.ALPHA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C(O)C(O)C1O[C@H](O)C(O)C1O"},
        "B_OCT": {"name": "Oct", "config": Config.BETA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC(O)C(O)C(O)C1O[C@@H](O)C(O)C1O"},

        # Pentose
        "PEN": {"name": "Pen", "config": Config.UNDEF, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                "smiles": "C(C1C(C(C(O1)O)O)O)O"},
        "A_PEN": {"name": "Pen", "config": Config.ALPHA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC1O[C@H](O)C(O)C1O"},
        "B_PEN": {"name": "Pen", "config": Config.BETA, "isomer": Enantiomer.U, "lactole": Lactole.FURANOSE,
                  "smiles": "OCC1O[C@@H](O)C(O)C1O"},

        # Glucose
        "GLC": {"name": "Glc", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]([C@@H]1[C@@H]([C@H](C(O1)O)O)O)O)O"},
        "A_GLC": {"name": "Glc", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@H](O)[C@H](O)[C@H]1O"},
        "B_GLC": {"name": "Glc", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@@H](O)[C@H](O)[C@H]1O"},

        # Mannose
        "MAN": {"name": "Man", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]([C@@H]1[C@@H]([C@@H](C(O1)O)O)O)O)O"},
        "A_MAN": {"name": "Man", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@H](O)[C@@H](O)[C@H]1O"},
        "B_MAN": {"name": "Man", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@@H](O)[C@@H](O)[C@H]1O"},

        # Galactose
        "GAL": {"name": "Gal", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]([C@H]1[C@@H]([C@H](C(O1)O)O)O)O)O"},
        "A_GAL": {"name": "Gal", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@H](O)[C@H](O)[C@H]1O"},
        "B_GAL": {"name": "Gal", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@@H](O)[C@H](O)[C@H]1O"},

        # Gulose
        "GUL": {"name": "Gul", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]([C@H]1[C@H]([C@H](C(O1)O)O)O)O)O"},
        "A_GUL": {"name": "Gul", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@H](O)[C@H](O)[C@@H]1O"},
        "B_GUL": {"name": "Gul", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@@H](O)[C@H](O)[C@@H]1O"},

        # Altrose
        "ALT": {"name": "Alt", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H](O)[C@@H]1OC(O)[C@H](O)[C@H]1O"},
        "A_ALT": {"name": "Alt", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H](O)[C@@H]1O[C@@H](O)[C@H](O)[C@H]1O"},
        "B_ALT": {"name": "Alt", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H](O)[C@@H]1O[C@H](O)[C@H](O)[C@H]1O"},

        # Allose
        "ALL": {"name": "All", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]([C@@H]1[C@H]([C@H](C(O1)O)O)O)O)O"},
        "A_ALL": {"name": "All", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@H](O)[C@H](O)[C@@H]1O"},
        "B_ALL": {"name": "All", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@@H](O)[C@H](O)[C@@H]1O"},

        # Talose
        "TAL": {"name": "Tal", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@@H](O)[C@@H]1OC(O)[C@@H](O)[C@H]1O"},
        "A_TAL": {"name": "Tal", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@H](O)[C@@H](O)[C@H]1O"},
        "B_TAL": {"name": "Tal", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@@H]1O[C@@H](O)[C@@H](O)[C@H]1O"},

        # Idose
        "IDO": {"name": "Ido", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H](O)[C@H]1OC(O)[C@H](O)[C@H]1O"},
        "A_IDO": {"name": "Ido", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H](O)[C@H]1O[C@@H](O)[C@H](O)[C@H]1O"},
        "B_IDO": {"name": "Ido", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H](O)[C@H]1O[C@H](O)[C@H](O)[C@H]1O"},

        # Quinovose
        "QUI": {"name": "Qui", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@@H](O)[C@H]1OC(O)[C@H](O)[C@H]1O"},
        "A_QUI": {"name": "Qui", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@H]1O[C@H](O)[C@H](O)[C@H]1O"},
        "B_QUI": {"name": "Qui", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@H]1O[C@@H](O)[C@H](O)[C@H]1O"},

        # Rhamnose
        "RHA": {"name": "Rha", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@H](O)[C@@H]1OC(O)[C@H](O)[C@@H]1O"},
        "A_RHA": {"name": "Rha", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@H](O)[C@@H]1O[C@@H](O)[C@H](O)[C@@H]1O"},
        "B_RHA": {"name": "Rha", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@H](O)[C@@H]1O[C@H](O)[C@H](O)[C@@H]1O"},

        # Fucose
        "FUC": {"name": "Fuc", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@H](O)[C@H]1OC(O)[C@@H](O)[C@@H]1O"},
        "A_FUC": {"name": "Fuc", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@H](O)[C@H]1O[C@@H](O)[C@@H](O)[C@@H]1O"},
        "B_FUC": {"name": "Fuc", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@H](O)[C@H]1O[C@H](O)[C@@H](O)[C@@H]1O"},

        # Olivose
        # Tyvelose

        # Abequose
        "ABE": {"name": "Abe", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@@H](O)[C@H]1C[C@@H](O)C(O)O1"},
        "A_ABE": {"name": "Abe", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@H]1C[C@@H](O)[C@H](O)O1"},
        "B_ABE": {"name": "Abe", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@H]1C[C@@H](O)[C@@H](O)O1"},

        # Paratose
        # Digitoxose
        # Colitose

        # Arabinose
        "ARA": {"name": "Ara", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]1[C@@H]([C@H](C(O1)O)O)O)O"},
        "A_ARA": {"name": "Ara", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H]1O[C@@H](O)[C@H](O)[C@H]1O"},
        "B_ARA": {"name": "Ara", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H]1O[C@H](O)[C@H](O)[C@H]1O"},

        # Lyxose
        "LYX": {"name": "Lyx", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)[C@@H](O)[C@H]1O"},
        "A_LYX": {"name": "Lyx", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@@H](O)[C@H]1O"},
        "B_LYX": {"name": "Lyx", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@@H](O)[C@H]1O"},

        # Xylose
        "XYL": {"name": "Xyl", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@H]1O"},
        "A_XYL": {"name": "Xyl", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@H]1O"},
        "B_XYL": {"name": "Xyl", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@H]1O"},

        # Ribose
        "RIB": {"name": "Rib", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)[C@H](O)[C@@H]1O"},
        "A_RIB": {"name": "Rib", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@H](O)[C@H](O)[C@@H]1O"},
        "B_RIB": {"name": "Rib", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@H](O)[C@H](O)[C@@H]1O"},

        # ketooctonic acid
        "KO": {"name": "Ko", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
               "smiles": "O=C(O)C1(O)O[C@@H]([C@H](O)[C@H](O)CO)[C@H](O)[C@@H]1O"},
        "A_KO": {"name": "Ko", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                 "smiles": "O=C(O)[C@@]1(O)O[C@@H]([C@H](O)[C@H](O)CO)[C@H](O)[C@@H]1O"},
        "B_KO": {"name": "Ko", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                 "smiles": "O=C(O)[C@]1(O)O[C@@H]([C@H](O)[C@H](O)CO)[C@H](O)[C@@H]1O"},

        # Keto-Deoxy-Nonulonic acid
        "KDN": {"name": "Kdn", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "O=C(O)C1(O)C[C@H](O)[C@H]([C@@H](O)[C@H](O)[C@H](O)CO)O1"},
        "A_KDN": {"name": "Kdn", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@@]1(O)C[C@H](O)[C@H]([C@@H](O)[C@H](O)[C@H](O)CO)O1"},
        "B_KDN": {"name": "Kdn", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@]1(O)C[C@H](O)[C@H]([C@@H](O)[C@H](O)[C@H](O)CO)O1"},

        # Neuraminic acid
        # Pseudaminic acid
        # Legionaminic acid
        # Acinetaminic acid
        # Bacillosamine

        # 2-Keto-3-Deoxy-D-Mannooctanoic acid
        "KDO": {"name": "Kdo", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "O=C(O)C1(O)C[C@@H](O)[C@H]([C@H](O)[C@H](O)CO)O1"},
        "A_KDO": {"name": "Kdo", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@@]1(O)C[C@@H](O)[C@H]([C@H](O)[C@H](O)CO)O1"},
        "B_KDO": {"name": "Kdo", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@]1(O)C[C@@H](O)[C@H]([C@H](O)[C@H](O)CO)O1"},

        # 3-Deoxy-D-Lyxo-Heptopyran-2-ularic acid
        "DHA": {"name": "Dha", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "O=C(O)[C@@H](O)[C@@H]1OC(O)(C(=O)O)C[C@H]1O"},
        "A_DHA": {"name": "Dha", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@@H](O)[C@@H]1O[C@@](O)(C(=O)O)C[C@H]1O"},
        "B_DHA": {"name": "Dha", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O=C(O)[C@@H](O)[C@@H]1O[C@](O)(C(=O)O)C[C@H]1O"},

        # Muramic acid
        "MUR": {"name": "Mur", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@@H](O[C@@H]1[C@@H](N)C(O)O[C@@H]1[C@H](O)CO)C(=O)O"},
        "A_MUR": {"name": "Mur", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O[C@@H]1[C@@H](N)[C@@H](O)O[C@@H]1[C@H](O)CO)C(=O)O"},
        "B_MUR": {"name": "Mur", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O[C@@H]1[C@@H](N)[C@H](O)O[C@@H]1[C@H](O)CO)C(=O)O"},

        # Apiose
        "API": {"name": "Api", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@@]1(O)COC(O)[C@@H]1O"},
        "A_API": {"name": "Api", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)CO[C@@H](O)[C@@H]1O"},
        "B_API": {"name": "Api", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)CO[C@H](O)[C@@H]1O"},

        # Fructose
        "FRU": {"name": "Fru", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)(CO)[C@@H](O)[C@@H]1O"},
        "A_FRU": {"name": "Fru", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@](O)(CO)[C@@H](O)[C@@H]1O"},
        "B_FRU": {"name": "Fru", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@](O)(CO)[C@@H](O)[C@@H]1O"},

        # Tagatose
        "TAG": {"name": "Tag", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)(CO)[C@@H](O)[C@H]1O"},
        "A_TAG": {"name": "Tag", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@](O)(CO)[C@@H](O)[C@H]1O"},
        "B_TAG": {"name": "Tag", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@](O)(CO)[C@@H](O)[C@H]1O"},

        # Sorbose
        "SOR": {"name": "Sor", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "C([C@H]1[C@H]([C@@H](C(O1)(CO)O)O)O)O"},
        "A_SOR": {"name": "Sor", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H]1O[C@](O)(CO)[C@@H](O)[C@@H]1O"},
        "B_SOR": {"name": "Sor", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H]1O[C@@](O)(CO)[C@@H](O)[C@@H]1O"},

        # Psicose
        "PSI": {"name": "Psi", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@H]1OC(O)(CO)[C@H](O)[C@@H]1O"},
        "A_PSI": {"name": "Psi", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@@](O)(CO)[C@H](O)[C@@H]1O"},
        "B_PSI": {"name": "Psi", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@H]1O[C@](O)(CO)[C@H](O)[C@@H]1O"},

        # Erythrose
        "ERY": {"name": "Ery", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "O[C@@H]1COC(O)[C@@H]1O"},
        "A_ERY": {"name": "Ery", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O[C@@H]1CO[C@H](O)[C@@H]1O"},
        "B_ERY": {"name": "Ery", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "O[C@@H]1CO[C@@H](O)[C@@H]1O"},

        # Threose
        "THRE": {"name": "Thre", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                 "smiles": "OC1OC[C@@H](O)[C@@H]1O"},
        "A_THRE": {"name": "Thre", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                   "smiles": "O[C@H]1OC[C@@H](O)[C@@H]1O"},
        "B_THRE": {"name": "Thre", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                   "smiles": "O[C@@H]1CO[C@@H](O)[C@H]1O"},

        # Ribulose
        "RUL": {"name": "Rul", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OCC1(O)OC[C@@H](O)[C@H]1O"},
        "A_RUL": {"name": "Rul", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@]1(O)OC[C@@H](O)[C@H]1O"},
        "B_RUL": {"name": "Rul", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@@H](O)[C@H]1O"},

        # Xylulose
        "XLU": {"name": "Xlu", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OCC1(O)OC[C@@H](O)[C@@H]1O"},
        "A_XLU": {"name": "Xlu", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@]1(O)OC[C@@H](O)[C@@H]1O"},
        "B_XLU": {"name": "Xlu", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@]1(O)OC[C@@H](O)[C@@H]1O"},

        # Aceric acid
        "ACE": {"name": "Ace", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@@H]1OC(O)[C@@H](O)[C@@]1(O)C(=O)O"},
        "A_ACE": {"name": "Ace", "config": Config.ALPHA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H]1O[C@@H](O)[C@@H](O)[C@@]1(O)C(=O)O"},
        "B_ACE": {"name": "Ace", "config": Config.BETA, "isomer": Enantiomer.L, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H]1O[C@H](O)[C@@H](O)[C@@]1(O)C(=O)O"},

        # Paratose
        "PAR": {"name": "Par", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "C[C@@H](O)[C@@H]1C[C@@H](O)C(O)O1"},
        "A_PAR": {"name": "Par", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@@H]1C[C@@H](O)[C@@H](O)O1"},
        "B_PAR": {"name": "Par", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "C[C@@H](O)[C@@H]1C[C@@H](O)[C@H](O)O1"},

        # Sedoheptulose
        "SED": {"name": "Sed", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                "smiles": "OC[C@@H](O)[C@H]1OC(O)(CO)[C@@H](O)[C@@H]1O"},
        "A_SED": {"name": "Sed", "config": Config.ALPHA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@@](O)(CO)[C@@H](O)[C@@H]1O"},
        "B_SED": {"name": "Sed", "config": Config.BETA, "isomer": Enantiomer.D, "lactole": Lactole.FURANOSE,
                  "smiles": "OC[C@@H](O)[C@H]1O[C@](O)(CO)[C@@H](O)[C@@H]1O"},
    }

    def __contains__(self, item):
        """
        Check if an item is part of this factory and can be returned.

        Args:
            item (str): Code of a monomer to be checked

        Returns:
            True if the item is included in the current version of this package
        """
        return item.upper() in FuranoseFactory.__monomers.keys()

    @staticmethod
    def keys():
        """
        Get all monomers that are included in this package by their extended name.

        Returns:
            Set of names for all furanoses (alpha/beta/undefined)
        """
        return FuranoseFactory.__monomers.keys()

    @staticmethod
    def monomers():
        """
        Get the names of all monomers in this package ignoring the alpha/beta conformations.

        Returns:
            List, sorted from long to short, of all monomer names in upper case
        """
        return list(set(x.split("_")[-1] for x in FuranoseFactory.__monomers.keys()))

    @staticmethod
    def __getitem__(item):
        """
        Get an instance of a monomer from this factory.

        Args:
            item (str): name of the query monomer

        Returns:
            Directory containing all necessary information to initialize a monomer implementation
        """
        return FuranoseFactory.__monomers[item.upper()]



import numpy as np
from rdkit.Chem import MolFromSmiles, MolToSmiles, GetAdjacencyMatrix

from glyles.glycans.enum_c import enumerate_carbon
from glyles.glycans.reactor import SMILESReaktor
from glyles.glycans.utils import Config, find_isomorphism_nx
from glyles.grammar.GlycanLexer import GlycanLexer


class Monomer:

    def __init__(self, origin=None, **kwargs):
        """
        Initialize the monomer using the super method. Additionally, some fields are initialized to describe the
        structure of the monomer according to the specification of the monomer-parent class

        Args:
            origin (Monomer): Other monomer to use to initialize this object
            **kwargs: arguments to initialize monomer if object is None.
        """
        if isinstance(origin, tuple):
            origin = origin[0]

        if origin is None:
            self.name = kwargs["name"]
            self.smiles = kwargs["smiles"]
            self.__root_smiles = kwargs["smiles"]
            self.structure = kwargs.get("struct", None)
            self.config = kwargs["config"]
            self.isomer = kwargs["isomer"]
            self.lactole = kwargs["lactole"]
            self.recipe = kwargs["recipe"]
            self.c1_find = kwargs.get("c1_find", None)
        else:
            self.name = origin.get_name()
            self.smiles = origin.get_smiles()
            self.__root_smiles = origin.root_smiles()
            self.c1_find = origin.get_c1_finder()
            self.structure = origin.get_structure()
            self.config = origin.get_config()
            self.isomer = origin.get_isomer()
            self.lactole = origin.get_lactole()
            self.recipe = origin.get_recipe()
            self.adjacency = origin.get_adjacency()
            self.ring_info = origin.get_ring_info()
            self.x = origin.get_features()

    def root_smiles(self):
        """
        Access the SMILES string of the root monomer. This is before any functional group is attached.
        It is as described in the factories.

        Returns:
            SMILES string of the root monomer of this monosaccharide
        """
        return self.__root_smiles

    def get_name(self, full=False):
        """
        Returns the name of this monomer as three-letter code (eventually longer for more fancy monosaccharides with
        more complex side chains).

        Args:
            full (bool): Flag indicating to also output recipe of monomer

        Returns:
            The name of this monomer
        """
        if full:
            return "".join((s for s, _ in self.recipe))
        return self.name

    def get_smiles(self):
        """
        Returns the SMILES representation of this monomer. Attention: These methods return the SMILES that is used to
        initialize this monomer. This is different from the to_smiles method of this class that returns the SMILES
        string with added place-holders that is used for the generation of the SMILES representation of the complete
        glycan

        Returns:
            The SMILES string that was used for initialization of this monomer
        """
        return self.smiles

    def get_c1_finder(self):
        """
        This returns a method used to identify the C1 atom in an additionally provided molecule object.

        Returns:
            The callable object that, given the structure, finds the RDKit ID of C1
        """
        return self.c1_find

    def alpha(self, factory):
        """
        Return this monosaccharide in its alpha conformation.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in alpha conformation
        """
        recipe = [(v, t) for v, t in self.recipe if t != GlycanLexer.TYPE]
        recipe.append(('a', GlycanLexer.TYPE))
        return Monomer(factory.create(recipe))

    def beta(self, factory):
        """
        Return this monosaccharide in its beta conformation.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in beta conformation
        """
        recipe = [(v, t) for v, t in self.recipe if t != GlycanLexer.TYPE]
        recipe.append(('b', GlycanLexer.TYPE))
        return Monomer(factory.create(recipe))

    def undefined(self, factory):
        """
        Return this monosaccharide in undefined conformation, the first carbon ring-atom will have unspecified.
        chirality.

        Args:
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            Monomer in undefined conformation
        """
        recipe = [(v, t) for v, t in self.recipe if t != GlycanLexer.TYPE]
        return Monomer(factory.create(recipe))

    def to_chirality(self, chirality, factory):
        """
        Return this monomer in the queried chirality.

        Args:
            chirality (str): char representing the chiral conformation of the first carbon ring atom
            factory (MonomerFactory): factory instance to be used to generate the monomers

        Returns:
            This monomer with the given (or not given) chirality at the first carbon ring atom
        """
        chirality = chirality.lower()
        if chirality == "a":
            return self.alpha(factory)
        if chirality == "b":
            return self.beta(factory)
        return self.undefined(factory)

    def get_config(self):
        """
        The current conformation relative to the first carbon ring-atom, i.e. alpha, beta or unspecified.

        Returns:
            Config-Tag according to the conformation this monomer represents
        """
        return self.config

    def get_isomer(self):
        """
        The current enantiomer of this monomer, i.e. L-form or D-form.

        Returns:
            Enantiomer-Tag according to the isomer this monomer represents
        """
        return self.isomer

    def get_lactole(self):
        """
        The current lactole-form of this monomer, i.e. if it's a 5-ring or a 6-ring molecule.

        Returns:
            Lactole-Tag according to the lactole-form this monomer represents
        """
        return self.lactole

    def get_recipe(self):
        """
        Return the recipe on how this monomer was created

        Returns:
            Recipe as list of modifications, confirmations and the root monomer as type List[Tuple[str, int]]
        """
        return self.recipe

    def is_non_chiral(self):
        """
        Check if this monomer represents a non-chiral molecule

        Returns:
            boolean indicating chirality of this monomer
        """
        return self.config == Config.UNDEF

    def get_adjacency(self):
        """
        Get the adjacency matrix of the atoms in this monomer.

        Returns:
            Adjacency matrix of all non-hydrogen atoms in this monomer
        """
        if self.adjacency is None:
            self.get_structure()
        return self.adjacency

    def get_ring_info(self):
        """
        Get information of all atom-ids in the rings of this monomer.

        Returns:
            Tuple of tuples with the atom-ids from rdkit in the monomer
        """
        if self.ring_info is None:
            self.get_structure()
        return self.ring_info

    def get_features(self):
        """
        Get a feature matrix from this monomer. The features contain information about atom type, ids according to the
        specification in monomer.py, and ring memberships.

        Returns:
            A numpy array of shape Nx3 containing the extracted features for all atoms in this molecule
        """
        if self.x is None:
            self.get_structure()
        return self.x

    @staticmethod
    def get_dummy_atoms():
        """
        Specify some dummy atoms that are used to mark oxygen atoms that will participate in bindings between glycans.
        Here, the atoms will be replaced by instances of the atom enum that are used to define the type of the atom in
        the nodes of the networkx representation of the monomer molecules.

        Returns:
            Two lists:
                * one with atoms that are given as the "atom" argument to the mark-method to replace the oxygen atoms
                * the string representation of the atoms from above, i.e. how the atoms above will be represented in a
                  SMILES string
        """
        return [34, 52, 84, 85], ["[SeH]", "[TeH]", "[PoH]", "[At]"]

    def root_atom_id(self, binding_c_id):
        """
        Get ID of atom that will bind the parent monomer in the glycan. This ID will be given as root argument to
        the to_smiles method.

        Args:
            binding_c_id (int): Integer at which c-position this monomer binds its parent

        Returns:
            id of the atom that binds to the parent, -1 if the root cannot be found
        """
        return self.find_oxygen(binding_c_id)

    def mark(self, position, atom):
        """
        Mark the oxygen atom linked to the carbon atom at the given position ready to participate in the bounding.
        Marking here works based on replacing the oxygen-group bound to the carbon atom at the given position with the
        atom enum instance also provided in the arguments.

        Args:
            position (int): id of the carbon atom whose oxygen atom will from the binding
            atom (int): atom to replace the binding oxygen with

        Returns:
            Nothing
        """
        idx = self.find_oxygen(position)
        self.get_structure().GetAtomWithIdx(idx).SetAtomicNum(atom)
        self.x[idx, 0] = atom

    def to_smiles(self, ring_index, root_idx=None, root_id=None):
        """
        Convert this monomer into a SMILES string representation.
        Use the implementation of the SMILES algorithm fitted to the needs of glycans.

        Args:
            ring_index (int): index of the rings in the atom
            root_idx (int): index of the root atom
            root_id (int): RDKit ID of root atom

        Returns:
            SMILES string representation of this molecule
        """
        assert root_idx is not None or root_id is not None, "Either Index or ID has to be provided"
        if root_id is None:
            if np.where(self.x[:, 1] == root_idx)[0].size != 0:
                root_id = int(np.where(self.x[:, 1] == root_idx)[0])
            else:
                root_id = int(np.where(self.x[:, 1] == 1)[0])
        smiles = MolToSmiles(self.get_structure(), rootedAtAtom=root_id)
        return "".join([((f"%{int(c) + ring_index}" if int(c) + ring_index >= 10
                          else f"{int(c) + ring_index}") if c.isdigit() else c) for c in smiles])

    def react(self, names, types):
        """
        Override the method to the call of the reactor to modify this monomer.

        Args:
            names (List[str]): name (string representation) of the modification
            types (List[int]): Type of the parsed stings based on GlycanLexer.TYPE

        Returns:
            New monomer with the altered structure
        """
        return SMILESReaktor(self).react(names, types)

    def get_structure(self):
        """
        Compute and save the structure of this glycan.

        Returns:
            rdkit molecule representing the structure of the glycan as a graph of its non-hydrogen atoms.
        """
        if self.structure is None:
            # read the structure from the SMILES string
            self.structure = MolFromSmiles(self.smiles)

            # extract some further information from the molecule to not operate always on the molecule
            self.adjacency = GetAdjacencyMatrix(self.structure, useBO=True)
            rings = self.structure.GetRingInfo().AtomRings()
            if len(rings) > 0 and self.name != "Inositol":
                self.ring_info = [None]
                for ring in rings:
                    found_ox = False
                    for atom in ring:
                        if self.structure.GetAtomWithIdx(atom).GetAtomicNum() == 8:
                            self.ring_info[0] = ring
                            found_ox = True
                            break
                    if not found_ox:
                        self.ring_info.append(ring)
            else:
                self.ring_info = rings
            self.x = np.zeros((self.adjacency.shape[0], 4), dtype=int)

            # extract some information form the molecule
            for i in range(self.adjacency.shape[0]):
                atom = self.structure.GetAtomWithIdx(i)

                # store the atom type
                self.x[i, 0] = atom.GetAtomicNum()

                # if the atom is part of any ring, store the number of that ring
                for r in range(len(self.ring_info)):
                    if self.ring_info[r] is not None and i in self.ring_info[r]:
                        self.x[i, 2] = r + 1

                # identify the oxygen atom in the main ring and set its id to 100
                if self.x[i, 2] == 1 and self.x[i, 0] == 8:
                    self.x[i, 1] = 100

            # identify isomorphic atoms. The reference for the isomorphism test is the root SMILES
            iso = list(find_isomorphism_nx(self.smiles, self.root_smiles(), self.name, self.c1_find).keys())
            if len(iso) == 0:
                iso = list(find_isomorphism_nx(self.root_smiles(), self.smiles, self.name, self.c1_find).values())
            if len(iso) == 0:
                self.x[:, 3] = 1
            else:
                self.x[iso, 3] = 1

            # Enumerate all carbons
            enumerate_carbon(self)

        return self.structure

    def find_oxygen(self, binding_c_id):
        """
        Find the oxygen atom that binds to the carbon atom with the provided id. The returned id may not refer to an
        oxygen atom in the ring of the monomer as this cannot bind anything. This method will report the atom id of the
        first atom type that fulfils the requirements of this method, i.e. the binding atom has to be the only one of
        its type, not in the main-ring, and connected to the given carbon atom.

        Args:
            binding_c_id (int): id of the carbon atom that participates in a binding, and we need to find the oxygen
                from

        Returns:
            The RDKit-ID referring to the atom binding the provided carbon atom and may participate in a glycan-binding.
            In case the carbon is bound to neither an oxygen nor a nitrogen, the RDKit-ID of the carbon is returned.
        """
        # first find the rdkit id of the carbon atom that should bind to something
        position = np.argwhere(self.x[:, 1] == binding_c_id).squeeze()

        multiple = False
        for check in [8, 7]:
            # then find the candidates. There should be exactly one element in the resulting array
            candidates = np.argwhere((self.adjacency[position, :] == 1) &
                                     (self.x[:, 0] == check) & (self.x[:, 2] != 1)).squeeze()
            if candidates.size == 1:
                return int(candidates)
            elif candidates.size > 0:
                multiple = True

        if not multiple and position.size == 1:
            return int(position)

        raise ValueError(f"Multiple options for oxygen (or other atom type) found.")



from rdkit import Chem
from rdkit.Chem import GetAdjacencyMatrix
from rdkit.Chem.rdchem import BondType

from glyles.glycans.utils import *
import numpy as np


class OpenFactory:
    """
    Checked with http://www.cheminfo.org/flavor/malaria/Utilities/SMILES_generator___checker/index.html
    """

    __monomers = {
        "HEX-OL": {"name": "Hexol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OCC(O)C(O)C(O)C(O)CO",
                   "c1_find": lambda x: c1_finder(x, "OCC(O)C(O)C(O)C(O)CO")},
        "GLC-OL": {"name": "Glucitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@H](O)[C@H](O)CO")},
        "MAN-OL": {"name": "Mannitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@@H](O)[C@@H](O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)[C@@H](O)[C@H](O)[C@H](O)CO")},
        "GAL-OL": {"name": "Galactitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@@H](O)[C@H](O)CO")},
        "GUL-OL": {"name": "Gulitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@H](O)[C@@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@H](O)[C@@H](O)[C@H](O)CO")},
        "ALT-OL": {"name": "Altritol", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@@H](O)[C@@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@@H](O)[C@@H](O)CO")},
        "ALL-OL": {"name": "Allitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@H](O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@H](O)[C@H](O)[C@H](O)CO")},
        "TAL-OL": {"name": "Talitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@@H](O)[C@@H](O)[C@@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)[C@@H](O)[C@@H](O)[C@H](O)CO")},
        "IDO-OL": {"name": "Iditol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@H](O)[C@@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@H](O)[C@@H](O)CO")},
        "QUI-OL": {"name": "Quinovitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@H](O)[C@H](O)C",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@H](O)[C@H](O)C")},
        "RHA-OL": {"name": "Rhamnitol", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@H](O)[C@@H](O)[C@@H](O)C",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@H](O)[C@@H](O)[C@@H](O)C")},
        "FUC-OL": {"name": "Fucitol", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@@H](O)[C@H](O)[C@H](O)[C@@H](O)C",
                   "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)[C@H](O)[C@H](O)[C@@H](O)C")},
        "ARA-OL": {"name": "Arabinitol", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@@H](O)CO")},
        "LYX-OL": {"name": "Lyxitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@@H](O)[C@@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)[C@@H](O)[C@H](O)CO")},
        "XYL-OL": {"name": "Xylitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@@H](O)[C@H](O)CO")},
        "RIB-OL": {"name": "Ribitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@H](O)[C@H](O)CO")},
        "KDO-OL": {"name": "Kdo alditol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC(=O)C(O)C[C@@H](O)[C@@H](O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC(=O)C(O)C[C@@H](O)[C@@H](O)[C@H](O)[C@H](O)CO")},
        "MUR-OL": {"name": "Muramic alditol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](N)[C@@H](O[C@H](C)C(=O)O)[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](N)[C@@H](O[C@H](C)C(=O)O)[C@H](O)[C@H](O)CO")},
        "API-OL": {"name": "Apiol", "config": Config.UNDEF, "isomer": Enantiomer.L, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@@H](O)C(O)(CO)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)C(O)(CO)CO")},
        "ERY-OL": {"name": "Erythriol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                   "smiles": "OC[C@H](O)[C@H](O)CO",
                   "c1_find": lambda x: c1_finder(x, "OC[C@H](O)[C@H](O)CO")},
        "THRE-OL": {"name": "Threitol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.OPEN,
                    "smiles": "OC[C@@H](O)[C@H](O)CO",
                    "c1_find": lambda x: c1_finder(x, "OC[C@@H](O)[C@H](O)CO")},
        "INS": {"name": "Inositol", "config": Config.UNDEF, "isomer": Enantiomer.D, "lactole": Lactole.PYRANOSE,
                "smiles": "[C@H]1(O)[C@@H](O)[C@@H](O)[C@H](O)[C@@H](O)[C@H](O)1",
                "c1_find": lambda x: c1_ino_finder(x)},
    }

    def __contains__(self, item):
        """
        Check if an item is part of this factory and can be returned.

        Args:
            item (str): Code of a monomer to be checked

        Returns:
            True if the item is included in the current version of this package
        """
        return item.upper() in OpenFactory.__monomers.keys()

    @staticmethod
    def __getitem__(item):
        """
        Get an instance of a monomer from this factory.

        Args:
            item (str): name of the query monomer

        Returns:
            Directory containing all necessary information to initialize a monomer implementation
        """
        return OpenFactory.__monomers[item.upper()]


def c1_ino_finder(structure):
    """
    For inositol, shorten that c1-finding process as this is only used for isomorphisms,
    return the list of carbon atoms.

    Args:
        structure (rdkit.Molecule): Molecule to determine the longest carbon-chain for starting at C1

    Returns:
        List of RDKit IDs starting from C1 all the way down the longest carbon chain
    """
    a_type = np.array([a.GetAtomicNum() for a in structure.GetAtoms()])
    return list(np.where(a_type == 6)[0].tolist())


def c1_finder(structure, base_smiles):
    """
    Determine the c1 atom of monosaccharides in open form. As the "classical" method does not apply, we compare the
    structure to the SMILES string that is generated from C1.

    Args:
        structure (rdkit.Molecule): Molecule to determine the longest carbon-chain starting at C1
        base_smiles (str): SMILES string of the root monomer starting from the oxygen of C1

    Returns:
        List of RDKit IDs starting from C1 all the way down the longest carbon chain
    """
    # compute atom types and adjacency matrix of molecule for faster computations
    a_type = np.array([a.GetAtomicNum() for a in structure.GetAtoms()])
    adjacency = GetAdjacencyMatrix(structure, useBO=True)
    c_atoms = np.where(a_type == 6)[0].tolist()

    # create a tree of all carbon atoms directly connected to the main ring of the monomer
    c_tree = Tree()
    stack = [(-1, c_atoms[0])]
    while len(stack) != 0:
        p_id, c_id = stack[-1]
        stack = stack[:-1]
        c_tree.add_node(c_id, p_id)

        children = np.argwhere((adjacency[c_id, :] == 1) & (a_type == 6))
        for c in children:
            if int(c) not in c_tree.nodes:
                stack.append((c_id, int(c)))

    # find the deepest node and rehang the tree to this node
    deepest_id, _ = c_tree.deepest_node()
    c_tree = c_tree.rehang_tree(deepest_id)
    longest_c_chain = c_tree.longest_chain()

    # now the two C1 candidates can be found at the ends of the longest chain
    start, end = longest_c_chain[0], longest_c_chain[-1]

    # check if exactly one of the two C1 candidates has a carboxyl group
    start_acid = check_for_acid(start, structure, adjacency, a_type)
    end_acid = check_for_acid(end, structure, adjacency, a_type)
    if start_acid and not end_acid:
        return longest_c_chain
    if end_acid and not start_acid:
        return reversed(longest_c_chain)

    # check if exactly one of the two C1 candidates has an attached oxygen
    start_aldehyd, end_aldehyd = check_for_aldehyd(start, adjacency, a_type), check_for_aldehyd(end, adjacency, a_type)
    if start_aldehyd and not end_aldehyd:
        return longest_c_chain
    if end_aldehyd and not start_aldehyd:
        return reversed(longest_c_chain)

    # compare chiral tags of molecule with SMILES string of root monomer. THIS IS INACCURATE DUE TO CHIRAL TAGS
    tmp = Chem.MolFromSmiles(base_smiles)
    tmp_chain = [a.GetIdx() for a in tmp.GetAtoms() if a.GetAtomicNum() == 6]
    for c, t in zip(longest_c_chain, tmp_chain):
        if structure.GetAtomWithIdx(int(c)).GetChiralTag() != tmp.GetAtomWithIdx(int(t)).GetChiralTag():
            return reversed(longest_c_chain)
    return longest_c_chain


def check_for_acid(start, structure, adjacency, a_type):
    """
    Check if the atom with the start ID has an acid group.

    Args:
        start (int): RDKit ID of the candidate for C1
        structure (rdkit.Molecule): rdkit molecule object
        adjacency (np.ndarray): adjacency matrix of the molecule
        a_type (np.ndarray): array of the atom types of the molecule's heavy atoms

    Returns:
        bool flag indicating if the carbon atom has an acid group
    """
    oxys = [int(x) for x in np.where((adjacency[start] != 0) & (a_type == 8))[0]]
    if len(oxys) != 2:
        return False
    if (structure.GetBondBetweenAtoms(int(start), oxys[0]).GetBondType() == BondType.SINGLE and
        structure.GetBondBetweenAtoms(int(start), oxys[1]).GetBondType() == BondType.DOUBLE) or \
            (structure.GetBondBetweenAtoms(int(start), oxys[1]).GetBondType() == BondType.SINGLE and
             structure.GetBondBetweenAtoms(int(start), oxys[0]).GetBondType() == BondType.DOUBLE):
        return True
    return False


def check_for_aldehyd(start, adjacency, a_type):
    """
    Check if the atom with the start ID has an aldehyd group.

    Args:
        start (int): RDKit ID of the candidate for C1
        adjacency (np.ndarray): adjacency matrix of the molecule
        a_type (np.ndarray): array of the atom types of the molecule's heavy atoms

    Returns:
        bool flag indicating if the carbon atom has an aldehyd group
    """
    return np.where((adjacency[start] == 1) & (a_type == 8))[0] != 0



from enum import Enum
from typing import Callable

import networkx as nx
from networkx.algorithms import isomorphism
from rdkit.Chem import MolFromSmiles, FindMolChiralCenters

from itertools import permutations


class Verbosity(Enum):
    """
    define Levels of verbosity for printing different important messages
    """
    QUIET = 0
    NORMAL = 1


class Config(Enum):
    """
    Configuration enum to represent if the monomer is in alpha, beta, or undefined configuration.
    """
    UNDEF = 0
    ALPHA = 1
    BETA = 2


class Enantiomer(Enum):
    """
    Configuration of the whole monomer regarding L and D forms in terms of enantiomerism.
    """
    D = 0
    L = 1
    U = 2


class Lactole(Enum):
    """
    Specification if a monomer is a pyranose (6-ring) or a furanose (5-ring)
    """
    UNKNOWN = 0
    OPEN = 1
    FURANOSE = 5
    PYRANOSE = 6


class UnreachableError(NotImplementedError):
    """
    Represent exceptions that should arise in case a piece of code is reached that under normal circumstances should
    not be reached
    """
    pass


class ParseError(ValueError):
    """
    Represent parsing errors when reading in a glycan
    """
    pass


ketoses2 = {
    ("Ko", Lactole.PYRANOSE), ("Kde", Lactole.PYRANOSE), ("Neu", Lactole.PYRANOSE), ("Pse", Lactole.PYRANOSE),
    ("Leg", Lactole.PYRANOSE), ("Aci", Lactole.PYRANOSE), ("Kdo", Lactole.PYRANOSE), ("Dha", Lactole.PYRANOSE),
    ("Fru", Lactole.PYRANOSE), ("Sor", Lactole.PYRANOSE), ("Tag", Lactole.PYRANOSE), ("Psi", Lactole.PYRANOSE),
}


def find_rings(mol):
    """
    Determine the rings of a molecule in RDKit.

    Args:
        mol (rdkit.Molecule): molecule to determine the rings for

    Returns:
        List (possibly empty) of rings of the input molecule. The list will contain a list of RDKit IDs of every atom
        in the respective rings
    """
    # get the rings from the molecule
    rings = mol.GetRingInfo().AtomRings()

    if len(rings) > 0:

        # if there are rings, reorder them to have the main-ring (with the oxygen) at first position
        ring_info = [None]
        for ring in rings:
            found_ox = False

            # if the ring contains an oxygen atom, it's the first/main ring
            for atom in ring:
                if mol.GetAtomWithIdx(atom).GetAtomicNum() == 8:
                    ring_info[0] = ring
                    found_ox = True
                    break

            # otherwise, just append it to the list of rings
            if not found_ox:
                ring_info.append(ring)

    # if there are no rings, return an empty list
    else:
        ring_info = rings

    return ring_info


def get_rings(mol, name, c1_find=None):
    """
    Get the longest carbon chain in the molecule starting from the C1 atom based on the rings of the molecule.

    Args:
        mol (rdkit.Molecule): molecule to determine the carbon list for
        name (str): Name of the monosaccharide we're working with
        c1_find (Callable): Optional function to determine C1 for non-ring monosaccharides

    Returns:
        List of the longest carbon chain starting with C1
    """
    # find the rings
    rings = find_rings(mol)
    if len(rings) == 0 or rings[0] is None or name == "Inositol":
        if c1_find is not None:
            return c1_find(mol)
        raise ValueError("Molecule should either have a ring or define a method to find c1!")

    # identify the oxygen in the first/main ring
    ox_id = -1
    for a in rings[0]:
        if mol.GetAtomWithIdx(a).GetAtomicNum() == 8:
            ox_id = a
            break
    if ox_id == -1:
        raise UnreachableError("First/Main should always have an oxygen atom in it!")

    # reorient the carbon list to start with the oxygen
    ring = rings[0][rings[0].index(ox_id):] + rings[0][:rings[0].index(ox_id)]

    return ring


def mol_to_nx(mol):
    """
    Convert a molecule to a networkx graph with atom types and bond types as only features.

    Args:
        mol (rdkit.Molecule): molecule to be converted

    Returns:
        networkx graph representing a molecule
    """
    g = nx.Graph()

    # add all the nodes/atoms
    for atom in mol.GetAtoms():
        g.add_node(atom.GetIdx(), type=atom.GetAtomicNum())

    # add all the bonds/edges
    for bond in mol.GetBonds():
        g.add_edge(bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(), type=bond.GetBondTypeAsDouble())
    return g


def networkx_match_rings(mol1_ring, mol2_ring):
    """
    Map the rings onto each other and yield every possible isomorphism between the rings

    Args:
        mol1_ring (networkx.Graph): graph representing the fist molecules ring
        mol2_ring (networkx.Graph): graph representing the second molecules ring

    Returns:
        Mappings between both rings corresponding to isomorphisms between both rings
    """
    # define a matcher for both graphs with the requirement matching all bond types and edge types for an isomorphism
    matcher = isomorphism.GraphMatcher(mol1_ring, mol2_ring, node_match=lambda n1, n2: n1["type"] == n2["type"],
                                       edge_match=lambda e1, e2: e1["type"] == e2["type"])

    # iterate over all possible isomorphism fulfilling previously mentioned requirement and yield them
    for mapping in matcher.subgraph_isomorphisms_iter():
        yield dict(mapping)


def networkx_fragment_isomorphism(mol1_nx, ring1, mol2_nx, ring2):
    """
    Find the biggest isomorphism between the two molecules represented in the graphs.

    Args:
        mol1_nx (networkx.Graph): graph of molecule1
        ring1 (List[int]): main-ring of molecule 1
        mol2_nx (networkx.Graph): graph of molecule 2
        ring2 (List[int]): ring of molecule 2

    Returns:
        Mapping from mol1's RDKIT IDs to mol2's RDKit IDs
    """
    # keep track of the biggest isomorphism between both molecules
    longest_iso = {}

    # create the graphs of the side chains (it's actually a collection of trees)
    mol1_no_ring = mol1_nx.subgraph(set(list(mol1_nx.nodes)).difference(set(ring1)))
    mol2_no_ring = mol2_nx.subgraph(set(list(mol2_nx.nodes)).difference(set(ring2)))

    # iterate over all isomorphisms between just the two rings
    for iso in networkx_match_rings(mol1_nx.subgraph(ring1), mol2_nx.subgraph(ring2)):
        ring_iso = {}

        # iterate over every matched position in the isomorphism
        for k, v in iso.items():

            # extract all neighbors at the given positions in the molecules
            k_neighbors = [x for x in mol1_nx.neighbors(k) if x not in ring1]
            v_neighbors = [x for x in mol2_nx.neighbors(v) if x not in ring2]

            # pad the neighborhood lists with Nones to have same length
            if len(k_neighbors) < len(v_neighbors):
                k_neighbors += [None for _ in range((len(v_neighbors) - len(k_neighbors)))]
            elif len(v_neighbors) < len(k_neighbors):
                v_neighbors += [None for _ in range((len(k_neighbors) - len(v_neighbors)))]

            pos_iso = {}

            # iterate over all permutation of the latter list...
            for v_neigh in permutations(v_neighbors):
                fg_iso = {}

                # ... to find the best isomorphism between the functional groups of Ck from mol1 and Cv from mol2
                for k_n, v_n in zip(k_neighbors, v_neigh):

                    # if one of the two side-chains is None, there is no isomorphism, so continue to next match
                    if k_n is None or v_n is None:
                        continue

                    pair_iso = {}

                    # define a matcher between the side chain trees ...
                    matcher = isomorphism.GraphMatcher(
                        mol1_nx.subgraph(nx.node_connected_component(mol1_no_ring, k_n).union([k])),
                        mol2_nx.subgraph(nx.node_connected_component(mol2_no_ring, v_n).union([v])),
                        node_match=lambda n1, n2: n1["type"] == n2["type"],
                        edge_match=lambda e1, e2: e1["type"] == e2["type"]
                    )

                    # ... and extract their best match
                    for mapping in matcher.subgraph_isomorphisms_iter():
                        if len(mapping) > len(pair_iso):
                            pair_iso = mapping

                    # put the best match of the side chains into the matching for the position isomorphisms
                    fg_iso.update(pair_iso)

                # keep track of the best isomorphisms between (potentially) different side chains at the same position
                if len(fg_iso) > len(pos_iso):
                    pos_iso = fg_iso

            # for each position, save the best isomorphism between the side chain trees
            ring_iso.update(pos_iso)

        # add the isomorphism of the side chains with the isomorphism of the ring
        ring_iso.update(iso)

        # save the best/biggest isomorphism between the two molecules
        if len(ring_iso) > len(longest_iso):
            longest_iso = ring_iso

    return longest_iso


def find_isomorphism_nx(mol1, mol2, name, c1_find=None):
    """
    Find an isomorphism between two molecules. The problem is an NP problem, normally. Here, we have some domain
    knowledge based on how and when this method is called.
    First, there is an isomorphism of at least the two rings. Second, the isomorphism can be extended by the attached
    functional groups. These functional groups are tree-like structured and not intersecting which makes it much easier
    to find isomorphisms between them.

    Args:
        mol1 (str): SMILES of the monosaccharide with functional groups
        mol2 (str): SMILES of the root monosaccharide
        name (str): Name of the monosaccharide we're working with
        c1_find (Callable): Optional method to provide to find C1 in special molecules

    Returns:
        Mapping from mol1's RDKIT IDs to mol2's RDKit IDs
    """
    # generate the RDKit molecules from the SMILES strings
    mol1_rd = MolFromSmiles(mol1)
    mol2_rd = MolFromSmiles(mol2)

    # identify the rings or the longest carbon-cain starting with
    ring1 = get_rings(mol1_rd, name, c1_find)
    ring2 = get_rings(mol2_rd, name, c1_find)

    # convert the molecules to networkx graphs for easier accessibility
    mol1_nx = mol_to_nx(mol1_rd)
    mol2_nx = mol_to_nx(mol2_rd)

    # actually compute the isomorphism between the two molecules and return it
    return networkx_fragment_isomorphism(mol1_nx, ring1, mol2_nx, ring2)


class Node:
    """
    Represent a node in a tree
    """

    def __init__(self, node_id, parent_id, depth, tree):
        """
        Initialize a node in of a tree
        Args:
            node_id (int): id of the node itself in the tree
            parent_id (int): id of the parent in the tree
            depth (int): depth of the node in the tree this node belongs to
            tree (Tree): reference to the tree this node belongs to
        """
        self.children = []
        self.node_id = node_id
        self.parent_id = parent_id
        self.depth = depth
        self.tree = tree

    def add_child(self, child_id):
        """
        Add a child node to this node

        Args:
            child_id (int): id of the added child node

        Returns:
            Nothing
        """
        self.children.append(child_id)

    def is_leaf(self):
        """
        Check if this node is a leaf

        Returns:
            true if this node has no children
        """
        return len(self.children) == 0

    def __str__(self):
        """
        Convert the node into a string representation

        Returns:
            String representation of this node and all its children
        """
        return "(" + str(self.node_id) + " [" + \
               ",".join([str(self.tree.nodes[child]) for child in self.children]) + "])"


class Tree:
    """
    Represent a tree with nodes
    """

    def __init__(self):
        """
        Initialize the tree as empty tree without any node
        """
        self.nodes = {}
        self.root = None

    def __str__(self):
        """
        Convert this tree into a string representation of all its nodes

        Returns:
            String representation of the tree seen from its root node
        """
        return str(self.nodes[self.root])

    def add_node(self, node_id, parent_id=-1):
        """
        Add a node to the tree

        Args:
            node_id (int): id of the new node
            parent_id (int): id of the parent of the new node

        Returns:
            Nothing
        """
        # check if the node is a root and the tree already has a root
        if parent_id == -1 and self.root is not None:
            raise ValueError("Tree cannot have two roots")
        if parent_id != -1 and parent_id not in self.nodes:
            raise ValueError("Parent node unknown")

        # add the node and eventually set the root to be that node
        if parent_id == -1:
            self.root = node_id
            self.nodes[node_id] = Node(node_id, parent_id, 0, self)
        else:
            self.nodes[node_id] = Node(node_id, parent_id, self.nodes[parent_id].depth + 1, self)

        # add the node to the list of children in its parent
        if parent_id != -1:
            self.nodes[parent_id].add_child(node_id)

    def deepest_node(self):
        """
        Find the deepest node in the tree (most distant to root)

        Returns:
            ID of the deepest node and its depth in the tree
        """
        deepest_id, deepest_depth = 0, 0
        for n_id, node in self.nodes.items():
            if node.depth > deepest_depth:
                deepest_depth = node.depth
                deepest_id = n_id
        return deepest_id, deepest_depth

    def rehang_tree(self, node_id):
        """
        Reorder the tree to start with the node with the given id in the root

        Args:
            node_id (int): is of the node to be used as new root node

        Returns:
            New tree with the specified node as root node
        """
        tree = Tree()
        stack = [(-1, node_id)]
        while len(stack) != 0:
            p_id, c_id = stack[-1]
            stack = stack[:-1]
            tree.add_node(c_id, p_id)

            children = self.nodes[c_id].children + [self.nodes[c_id].parent_id]
            for c in children:
                if c not in tree.nodes and c != -1:
                    stack.append((c_id, c))

        return tree

    def longest_chain(self, node_id=None):
        """
        Find the longest chain of nodes in the parent. This chain starts in the provided node and goes down in the
        tree. This might lead to false results of the node with the given id has two or more children.

        Args:
            node_id (int): id of the node to start in, if None, the root will be the start

        Returns:
            List of nodes along deepest path down the tree
        """
        if node_id is None:
            node_id = self.root

        if self.nodes[node_id].is_leaf():
            return [node_id]

        longest_chain = []
        for child in self.nodes[node_id].children:
            tmp = self.longest_chain(child)
            if len(tmp) > len(longest_chain):
                longest_chain = tmp

        return [node_id] + longest_chain



import copy
import logging
import re

import numpy as np
from rdkit.Chem.rdchem import ChiralType
from rdkit.Chem.rdmolops import AddHs, RemoveHs

from glyles.glycans.factory.factory_o import OpenFactory
from glyles.glycans.utils import Enantiomer, ketoses2
from glyles.grammar.GlycanLexer import GlycanLexer

O, C = 0, 1

# map of all functional groups from IUPAC name to SMILES string
functional_groups = {
    "": "",
    "Aep": "NCCP([O-])(=O)[O-]",
    "Ala": "N[C@@H](C)C(=O)O",
    "Am": "C(=N)C",
    "Asp": "N[C@@H](CC(=O)O)C(=O)O",
    "Br": "Br",
    "Cer": "OC[C@H](NC=O)C(O)/C=C/" + "C" * 13,
    "Cho": "OCC[N+](C)(C)C",
    "Cl": "Cl",
    "Cm": "NC(=O)",
    "Cys": "N[C@@H](CS)C(=O)O",
    "Etn": "OCCN",
    "EtN": "OCCN",
    "F": "F",
    "Glu": "OC(=O)CC[C@H](N)C(=O)O",
    "Hse": "OCCC(N)C(=O)O",
    "HSer": "OC(=O)[C@H](N)CCO",
    "I": "I",
    "Lys": "NCCCC[C@H](N)C(=O)O",
    "N": "N",
    "NFo": "NC=O",
    "Orn": "NCCC[C@H](N)C(=O)O",
    "P": "OP(=O)(O)O",
    "PhNO2": "Oc2ccc([N+]([O-])=O)cc2",
    "Pp": "OC(=O)CC",
    "S": "S(=O)(=O)O",
    "Pro": "N2CCCC2C(=O)O",
    "Ser": "OC(=O)[C@H](N)CO",
    "Tf": "OS(=O)(=O)C(F)(F)F",
    "Thr": "N[C@@H](C(=O)O)[C@@H](O)C",
    "Ulo": "OC(c2ccccc2)(c2ccccc2(Cl))CCN(C)C",

    # COH land
    "A": "OC(=O)C",
    "Allyl": "CC=C",
    "Ac": "C(=O)C",
    "Ang": "OC(=O)C/(C)=C\C",
    "Bz": "C(=O)c2ccccc2",
    "Bn": "OCc2ccccc2",
    "cdPam": "OC(=O)CCCCCCC/C=C\CCCCCC",
    "Cet": "CCC(=O)O",
    "Cin": "OC(=O)/C=C/c2ccccc2",
    "Dce": "OC(=O)CCCCCCCC=C",
    "Dhp": "OC(=O)C(O)(O)CCC",
    "Dhpa": "OC(=O)C(O)(O)CCC",
    "Etg": "OCCO",
    "Fer": "OC(=O)/C=C/c2ccc(O)c(OC)c2",
    "Fo": "OC(=O)",
    "Gc": "C(=O)CO",
    "Gly": "OCC(O)CO",
    "Gro": "OCC(O)CO",
    "He": "C(O)C",
    "Lac": "OC(=O)C(O)C",
    "Lin": "OC(=O)CCCCCCC/C=C\C/C=C\CCCCC",
    "Mal": "O[C@H](C(=O)O)CC(=O)O",
    "Ole": "OC(=O)CCCCCCC/C=C\CCCCCCCC",
    "Ph": "c2ccccc2",
    "Phyt": "OCCC(C)CCCC(C)CCCC(C)CCCC(C)C",
    "Pyr": "OC(=O)C(=O)C",
    "Sin": "OC(=O)/C=C/c2cc(OC)c(O)c(OC)c2",
    "Suc": "OC(=O)CCC(=O)O",
    "Tig": "OC(=O)C/(C)=C/C",
    "Tr": "C(c2ccccc2)(c3ccccc3)c4ccccc4",
    "Ts": "OS(=O)(=O)c2ccc(C)cc2",
    "Vac": "OC(=O)CCCCCCCCCC=CCCCCCC",

    # Sugar rings
    "Pen": "OC2C(O)C(O)C(O)CO2",
    "Hex": "OC2C(O)C(O)C(O)C(CO)O2",
    "Hep": "OC2C(O)C(O)C(O)C(C(O)CO)O2",
    "Oct": "OC2C(O)C(O)C(O)C(C(O)C(O)CO)O2",

    # some special cases
    "3oxoMyr": "OC(=O)CC(=O)CCCCCCCCCCC",
    "17HOLin": "OC(=O)CCCCCCC/C=C\C/C=C\CCCC(O)C",
    "aLnn": "OC(=O)CCCCCCCC=CCC=CCC=CCC",
    "cVac": "OC(=O)CCCCCCCCC/C=C\CCCCCC",
    "d2Ach": "OC(=O)CCCCCCC=CCC=CCCCCCCCC",
    "d3Ach": "OC(=O)CCCC=CCC=CCC=CCCCCCCCC",
    "d4Ach": "OC(=O)CCCC=CCC=CCC=CCC=CCCCCC",
    "dPam": "OC(=O)CCCCCCCC=CCCCCCC",
    "eSte": "OC(=O)CCCCCCCC=CC=CC=CCCCC",
    "gLnn": "OC(=O)CCCCC=CCC=CCC=CCCCCC",
    "tdPam": "OC(=O)CCCCCCC/C=C/CCCCCC",
    "tBu": "OC(C)(C)C",

    # Methanol, Ethanol, Propanol, ...
    "Me": "O" + "C" * 1,
    "Et": "O" + "C" * 2,
    "Pr": "O" + "C" * 3,
    "Prop": "O" + "C" * 3,
    "Bu": "O" + "C" * 4,
    "Pe": "O" + "C" * 5,
    "Hx": "O" + "C" * 6,
    "Hp": "O" + "C" * 7,
    "Oc": "O" + "C" * 8,
    "Nn": "O" + "C" * 9,
    "Dec": "O" + "C" * 10,
    "Und": "O" + "C" * 11,
    "Dod": "O" + "C" * 12,
    # ... and their acids
    "But": "OC(=O)" + "C" * 2,
    "Vl": "OC(=O)" + "C" * 4,
    "Hxo": "OC(=O)" + "C" * 5,
    "Hpo": "OC(=O)" + "C" * 6,
    "Oco": "OC(=O)" + "C" * 7,
    "Nno": "OC(=O)" + "C" * 8,
    "Non": "OC(=O)" + "C" * 8,
    "Dco": "OC(=O)" + "C" * 9,
    "Udo": "OC(=O)" + "C" * 10,
    "Lau": "OC(=O)" + "C" * 11,
    "Myr": "OC(=O)" + "C" * 13,
    "Pam": "OC(=O)" + "C" * 15,
    "Mar": "OC(=O)" + "C" * 16,
    "Ste": "OC(=O)" + "C" * 17,
    "Ach": "OC(=O)" + "C" * 19,
    "Hico": "OC(=O)" + "C" * 20,
    "Beh": "OC(=O)" + "C" * 21,
    "Lig": "OC(=O)" + "C" * 23,
    "Crt": "OC(=O)" + "C" * 25,
    "Ccr": "OC(=O)" + "C" * 26,
    "Mon": "OC(=O)" + "C" * 27,
    "Mel": "OC(=O)" + "C" * 28,
    "Lacceroic": "OC(=O)" + "C" * 31,
    "Psyllic": "OC(=O)" + "C" * 32,
    "Geddic": "OC(=O)" + "C" * 33,
    "Ceroplastic": "OC(=O)" + "C" * 35,
}

# list of functional groups that preserve the atom it is attached to instead of replacing it
preserve_elem = [
    "Ac", "Allyl", "Am", "Bz", "Gc", "P", "S"
]

# list of functional groups that start with an N and might be confused with a nitrogen-bridge
n_conflict = [
    "Nno", "Non", "Nn"
]

# list of functional groups that start with an O and might be confused with an oxygen-bridge
o_conflict = [
    "Oco", "Ole", "Orn", "Oc"
]

# list of functional groups that start with a P and might be confused with a phosphate-bridge
p_conflict = [
    "Psyllic", "Prop", "Pam", "Pro", "Pyr", "Pe", "Ph", "Pr", "Pp"
]

# list of functional groups that start with a C and might be confused with a carbon-bridge
c_conflict = [
    "Cct", "Cer", "Cet", "Cho", "Cin", "Crt", "Cys", "Cl", "Cm"
]


def not_implemented_message(mod):
    """
    Print a message if some modification is parsable but not implemented yet. So, the grammar might accept some
    modifications but the reactor is yet not capable to attach that functional group to the monomer

    Args
        mod (str): name of the modification that cannot be converted

    Returns
        Nothing
    """
    logging.warning(
        f"ModificationNotImplementedWarning: {mod} Modification not implemented. "
        f"The returned molecule will not have this modification."
    )


def extract_bridge(n):
    """
    Extract bridge and functional groups from the IUPAC description.

    Args:
        n (str): name of functional group in IUPAC format

    Returns:
        Tuple of bridge element and the functional group
    """
    # Find the lowercase part that is for sure part of the functional group's name
    i = re.search('[a-z]', n)
    if i is not None:
        i = i.start()
        if i >= 2 and n[i - 2] == "H":  # in case of something like Homoserine (HSer)
            i -= 1
        fg = n[i - 1:]
        bridge = n[:i - 1]

        # if bridge overlaps with functional group
        bridge = bridge[:-1] if functional_groups[fg][0] == bridge[-1] else bridge
    else:
        # if no lowercase in input the last char is the functional group, everything before the bridge
        fg = n[-1]
        bridge = n[:-1]

    # if bridge starts with a number, drop it
    if len(bridge) > 0 and bridge[0].isdigit():
        bridge = bridge[1:]

    # if bridge starts with carbon, drop it
    if len(bridge) > 0 and bridge[0] == "C":
        bridge = bridge[1:]

    return bridge, fg


def opposite_chirality(tag):
    """
    Invert chriality based on the given ChiralType

    Args:
        tag (rdkit.Chem.rdchem.ChiralType): ChiralType to be inverted

    Returns:
        Returns opposite chiral tag
    """
    if tag == ChiralType.CHI_TETRAHEDRAL_CCW:
        return ChiralType.CHI_TETRAHEDRAL_CW
    elif tag == ChiralType.CHI_TETRAHEDRAL_CW:
        return ChiralType.CHI_TETRAHEDRAL_CCW
    return tag


class SMILESReaktor:
    def __init__(self, monomer):
        """
        Initialize the reaktor with the monosaccharide to attach functional groups to
        Args:
            monomer (Monomer): monosaccharide object
        """
        self.monomer = monomer
        self.monomer.get_structure()
        self.side_chains = None
        self.ring_c = 2 if (monomer, monomer.get_lactole()) in ketoses2 else 1

    def react(self, names, types):
        """
        Main function to add functional groups to the monomer stored in this object.

        Args:
            names (List[str]): List of the names of the functional groups
            types (List[int]): List of type definitions from the recipe of a monosaccharide

        Returns:
            The monomer object and a bool flag indicating that all functional groups were recognized and attached
        """
        full = True

        # parse for sth. like LDManHep or DDManHep -> afterwards, no need to look for LD/DD/... and Hep/Hex/Pen/Oct
        self.check_for_resizing(names, types)

        # check for open forms like "-ol" and "-onic"
        self.check_for_open_form(names, types)

        # store all functional groups that cannot be attached in the current round for the next round
        higher_order_groups = [], []
        while len(names) > 0:
            start_len = len(names)

            # define list of functional groups to be attached
            self.side_chains = [["", ""] for _ in range(1 + np.count_nonzero(self.monomer.x[:, 0] == 6))]

            # parse remaining modifications
            for n, t in zip(names, types):
                # if it's not a modification or already parsed, continue
                if t != GlycanLexer.MOD or n.count("L") + n.count("D") == len(n) or n in ['-', '-ol', '-onic']:
                    continue

                # if the name starts with a - char, drop this, the functional group will be independent of the previous
                if n[0] == "-":
                    n = n[1:]

                # parse making the monosaccharide an acid
                if n == "A":
                    # if there's no ring take carbon with the highest number
                    if sum(self.monomer.x[:, 2] == 1) == 0:
                        c_id = int(max(self.monomer.x[self.monomer.x[:, 0] == 6, 1]))

                    # else take the last carbon in the ring
                    else:
                        c_id = int(max(self.monomer.x[(self.monomer.x[:, 0] == 6) & (self.monomer.x[:, 2] == 1), 1]))
                        c_id = int(np.where(self.monomer.x[:, 1] == c_id)[0])

                    # if the selected carbon has a tail raging away from the monomer, iterate all the way down
                    children = np.where((self.monomer.adjacency[c_id, :] == 1) & (self.monomer.x[:, 0] == 6) &
                                        (self.monomer.x[:, 2] == 0))[0].tolist()
                    while len(children) != 0:
                        c_id = int(children[0])
                        children = np.where(
                            (self.monomer.adjacency[c_id, :] == 1) & (self.monomer.x[:, 0] == 6) &
                            (self.monomer.x[:, 2] == 0) & (self.monomer.x[:, 1] > self.monomer.x[c_id, 1])
                        )[0].tolist()

                    # attach that acid group in one way or another
                    if len(self.side_chains[self.monomer.x[c_id, 1]][O]) > 0 and \
                            self.side_chains[self.monomer.x[c_id, 1]][O][-1] == "O":
                        self.side_chains[self.monomer.x[c_id, 1]][O] += "C(=O)O"
                    else:
                        self.side_chains[self.monomer.x[c_id, 1]][O] += "(=O)O"

                # add a nitrogen
                elif n == "N":
                    self.side_chains[1 if self.monomer.get_name() in ['Fru', 'Tag', 'Sor', 'Psi'] else 2][O] += "N"

                # change the enantiomer to the D or L form
                elif n == "D-":
                    self.to_enantiomer(Enantiomer.D)
                elif n == "L-":
                    self.to_enantiomer(Enantiomer.L)

                # add acid or glycoly groups to neuraminic acid
                elif n == "Ac" and self.monomer.get_name() == "Neu":
                    self.side_chains[5][O] += "NC(=O)C"
                elif n == "Gc" and self.monomer.get_name() == "Neu":
                    self.side_chains[5][O] += "NC(=O)CO"

                # if the side chain starts with a position specification
                elif n[0].isdigit():

                    # if the functional group cannot be attached at the moment, wait for next round
                    if int(n[0]) > len(self.side_chains) - 1:
                        higher_order_groups[0].append(n)
                        higher_order_groups[1].append(t)
                        continue

                    # desoxygenation of a specific position
                    if n[1:] == "d":
                        self.side_chains[int(n[0])][O] += "H"

                    # change chirality at a single chiral carbon atom
                    elif n[1:] == "e":
                        idx = int(np.where(self.monomer.x[:, 1] == int(n[0]))[0])
                        tag = opposite_chirality(self.monomer.structure.GetAtomWithIdx(idx).GetChiralTag())
                        self.monomer.structure.GetAtomWithIdx(idx).SetChiralTag(tag)

                    # add a functional group connected with an oxygen or nitrogen
                    elif len(n) > 4 and n[1] == n[3] == "-" and n[2] in "ON":
                        elem = "" if functional_groups[n[4:-1]][0] == n[2] else n[2]
                        full &= self.set_fg(O, int(n[0]), elem, n[4:-1])

                    # connect a functional group with nitrogen, oxygen, or phosphate in between
                    elif n[1] in "NOP" and n[1:] not in n_conflict + o_conflict + p_conflict:
                        bridge, fg = extract_bridge(n)
                        if len(bridge) > 0 and functional_groups[fg][0] == bridge[-1]:
                            bridge = bridge[:-1]
                        full &= self.set_fg(O, int(n[0]), bridge, fg)

                    # add a group connected directly to the C-Atom
                    elif n[1] == "C" and n[1:] not in c_conflict:
                        bridge, fg = extract_bridge(n)
                        full &= self.set_fg(C, int(n[0]), bridge, fg)

                    # otherwise just add the sidechain directly to the specified position
                    else:
                        elem = self.monomer.structure.GetAtomWithIdx(self.monomer.find_oxygen(int(n[0]))).GetSymbol() \
                            if n[1:] in preserve_elem else ""
                        elem = "" if elem == "C" else elem
                        full &= self.set_fg(O, int(n[0]), elem, n[1:])

                # if the side-chain is connected with a nitrogen/oxygen/phosphate to the monomer
                elif n[0] in "NOP" and n not in n_conflict + o_conflict + p_conflict:
                    bridge, fg = extract_bridge(n)
                    if len(bridge) > 0 and functional_groups[fg][0] == bridge[-1]:
                        bridge = bridge[:-1]

                    # attach the monomer to the second carbon in the ring, it might be C2 or C3 (for 2-ketoses)
                    if fg == "Me":
                        full &= self.set_fg(O, self.ring_c, bridge, fg)
                    else:
                        full &= self.set_fg(O, self.ring_c + 1, bridge, fg)

                # functional groups might be directly connected to the carbon of the ring
                elif n[0] == "C" and n not in c_conflict:
                    if "=" in n or n[1:].isdigit():
                        self.set_fg(O, self.ring_c, "O", self.parse_poly_carbon(n))
                    else:  # add a group connected directly to the C-Atom
                        bridge, fg = extract_bridge(n)
                        full &= self.set_fg(C, self.ring_c, bridge, fg)

                # in all other cases just add the functional group
                else:
                    elem = self.monomer.structure.GetAtomWithIdx(self.monomer.find_oxygen(int(n[0]))).GetSymbol() \
                        if n[1:] in preserve_elem else ""
                    elem = "" if functional_groups[n][0] == elem else elem
                    full &= self.set_fg(O, self.ring_c, elem, n)

            # after storing all functional groups in a list, iterate over them and put them into the monomers structure
            self.assemble_chains()

            if len(higher_order_groups[0]) == start_len:
                break

            names = higher_order_groups[0]
            types = higher_order_groups[1]
            higher_order_groups = [], []

        return self.monomer, full and len(higher_order_groups[0]) == 0

    def set_fg(self, c_or_o, pos, bond_elem, name):
        """
        Add the recognized functional group to the array storing all functional groups.

        Args:
            c_or_o (int): Integer indicating if the functional group is attached to an oxygen or to the carbon directly
            pos (int): The position where to attach the group
            bond_elem (str): the element(s) to use for bridging between the functional group and the monomer
            name (str): THe IUPAC-condensed name of the functional group

        Returns:
            bool flag indicating that the functional group can be attached
        """
        # check if the group is known ...
        if name in functional_groups:
            if len(self.side_chains[pos][c_or_o]) > 0:
                if self.side_chains[pos][c_or_o][-1] == bond_elem:
                    bond_elem = ""
            if len(self.side_chains[pos][c_or_o]) > 0 and \
                    self.side_chains[pos][c_or_o][-1] == functional_groups[name][0] != "C":
                self.side_chains[pos][c_or_o] += bond_elem + functional_groups[name][1:]
            else:
                self.side_chains[pos][c_or_o] += bond_elem + functional_groups[name]
            return True

        # otherwise print message and return accordingly
        else:
            not_implemented_message(name)
            return False

    def check_for_resizing(self, names, types):
        """
        Change the monosaccharide in case it's indicated that the monosaccharide has more carbon atoms than normal.

        Args:
            names (List[str]): List of the names of the functional groups
            types (List[int]): List of type definitions from the recipe of a monosaccharide

        Returns:
            Nothing
        """
        sac_index = types.index(GlycanLexer.SAC)

        # if there's nothing to do, return
        if len(types) == sac_index + 1 or (len(types) > sac_index + 1 and types[sac_index + 1] != GlycanLexer.SAC):
            return

        # identify the root and the new size of the monomer by identifying the indices of their descriptions
        if names[sac_index + 1] in ["Pen", "Hex", "Hep", "Oct"]:
            len_index = sac_index + 1
        else:
            len_index = sac_index
            sac_index = len_index + 1

        # extract the orientations of the hydroxy-groups along the enlarged chain
        if sac_index != 0 and names[sac_index - 1].count("L") + \
                names[sac_index - 1].count("D") == len(names[sac_index - 1]):
            orientations = names[sac_index - 1]
        else:
            orientations = ""

        # determine the chain that has to be added to the last carbon atom
        c_count = np.count_nonzero(self.monomer.x[:, 0] == 6)
        if names[len_index] == "Pen":
            count = 5
        elif names[len_index] == "Hex":
            count = 6
        elif names[len_index] == "Hep":
            count = 7
        elif names[len_index] == "Oct":
            count = 8
        else:
            count = c_count
        extension = "[C?H](O)" * (count - c_count) + "CO"

        # insert the orientations
        for c in orientations[:-1]:
            if c == "L":
                extension = extension.replace("[C?H]", "[C@@H]", 1)
            if c == "D":
                extension = extension.replace("[C?H]", "[C@H]", 1)
        extension = extension.replace("[C?H]", "C")

        # find oxygen of c6 to delete it and find c6 to replace it
        c_id = int(np.where(self.monomer.x[:, 1] == int(max(self.monomer.x[self.monomer.x[:, 0] == 6, 1])))[0])
        # if the carbon to be extended is part of the ring, put the extension into a side_chain
        if self.monomer.x[c_id, 2] == 1:
            extension = extension.replace(")", ")(", 1) + ")"
        # if the original molecule has no oxygen at its last carbon, remove the first oxygen from the extension
        if self.monomer.x[((self.monomer.adjacency[c_id] - self.monomer.x[:, 2]) > 0) &
                          (self.monomer.x[:, 0] == 8), 0].size == 0:
            extension = extension.replace("(O)", "", 1)
        self.monomer.structure.GetAtomWithIdx(self.monomer.find_oxygen(c_count)).SetAtomicNum(50)
        self.monomer.structure.GetAtomWithIdx(c_id).SetAtomicNum(32)

        # generate SMILES from rings oxygen and replace c6's oxygen and c6 by the extension
        smiles = self.monomer.to_smiles(0, root_idx=10)
        smiles = smiles.replace("[SnH]", "").replace("[GeH2]", extension).replace("[GeH]", extension).replace("()", "")

        # update the monomer
        self.monomer.smiles = smiles
        self.monomer.structure = None
        self.monomer.get_structure()

    def check_for_open_form(self, names, types):
        """
        This method checks if the monomer is converted into an open form.

        Args:
            names (List[str]): List of the names of the functional groups
            types (List[int]): List of type definitions from the recipe of a monosaccharide

        Returns:
            Nothing
        """
        # stop if nothing to do here or the instructions are ambiguous
        if not (("-ol" in names) ^ ("-onic" in names)):
            return

        # identify the descriptions of what to do and determine the new SMILES string
        sac_index = types.index(GlycanLexer.SAC)
        params = copy.copy(OpenFactory()[names[sac_index] + "-ol"])
        if "-onic" in names:
            params["smiles"] = params["smiles"].replace("C", "C(=O)", 1)

        # update the monomer accordingly
        self.monomer.smiles = params["smiles"]
        self.monomer.c1_find = params["c1_find"]
        self.monomer.structure = None
        self.monomer.get_structure()

    def add_to_oxygen(self, chain, idx, placeholder):
        """
        Add a functional group to an oxygen.

        Args:
            chain (str): functional group to be attached
            idx (int): position where to bind the functional group
            placeholder (int): placeholder atom to be inserted at that position to replace it later in the SMILES

        Returns:
            Nothing
        """
        if chain == "H":  # account for desoxygenation
            self.monomer.structure.GetAtomWithIdx(idx).SetAtomicNum(1)
        else:
            self.monomer.structure.GetAtomWithIdx(idx).SetAtomicNum(placeholder)

    def add_to_carbon(self, chain, i, placeholder):
        """
        Add a functional group directly to a carbon.

        Args:
            chain (str): functional group to be attached
            i (int): position where to bind the functional group
            placeholder (int): placeholder atom to be inserted at that position to replace it later in the SMILES

        Returns:
            Nothing
        """
        # make hydrogens visible in order to replace one of them later
        tmp = AddHs(self.monomer.structure)

        # identify a hydrogen bound to the carbon atom to attach the functional group to
        h = None
        for n in tmp.GetAtomWithIdx(int(np.where(self.monomer.x[:, 1] == i)[0])).GetNeighbors():
            if n.GetAtomicNum() == 1:
                h = n.GetIdx()
                break
        # if none were found, raise an error
        if h is None:
            raise ValueError(f"There is no oxygen, nitrogen, and hydrogen attached to C{i}! "
                             f"No functional group can be attached there!")

        # replace the hydrogen with the functional groups placeholder
        tmp.GetAtomWithIdx(h).SetAtomicNum(placeholder)
        self.monomer.structure = RemoveHs(tmp)
        if chain[0] == "O" and chain not in preserve_elem:
            self.side_chains[i][C] = self.side_chains[i][C][1:]

    def assemble_chains(self):
        """
        Attach the chains to the monosaccharide.

        Returns:
            Nothing
        """
        # if there is no functional group to attach, return
        if not "".join("".join(x) for x in self.side_chains):
            return

        # TODO: Add more phs to be able to have both, carbon and oxygen attached functional groups at the same position
        placeholder = [
            (31, "[GaH2]"), (32, "[GeH3]"),
            (49, "[InH2]"), (50, "[SnH]"), (51, "[SbH2]"), (52, "[TeH]"),
            (81, "[TlH2]"), (82, "[PbH]"), (83, "[BiH2]"), (84, "[PoH]"),
        ]

        # iterate over all functional groups, ...
        for i, (chain, c_chain) in enumerate(self.side_chains):
            # ... check if they exist and in case add the groups
            if chain:
                idx = self.monomer.find_oxygen(i)
                if self.monomer.structure.GetAtomWithIdx(idx).GetSymbol() != "C":
                    self.add_to_oxygen(chain, idx, placeholder[i][0])
                else:
                    self.add_to_carbon(chain, i, placeholder[i][0])
            if c_chain:
                self.add_to_carbon(c_chain, i, placeholder[i][0])

        # create SMILES from molecule with placeholders
        smiles = self.monomer.to_smiles(0, root_idx=100)

        # replace all placeholders by their actual functional group
        for i, (chain, c_chain) in enumerate(self.side_chains):
            if chain:
                smiles = smiles.replace(placeholder[i][1], "" if chain == "H" else chain)

        # update the monomer accordingly
        self.monomer.smiles = smiles.replace("()", "")
        self.monomer.structure = None
        self.monomer.get_structure()

    def to_enantiomer(self, form):
        """
        Convert the stored monomer into the specified enantiomer by switching all chiral tags in the rings carbon atoms.

        Args:
            form (Enantiomer): Enantiomer, specifying L or D form

        Returns:
            Nothing
        """
        # check if the enantiomeric requirement is not already fulfilled
        if self.monomer.get_isomer() == form:
            return

        # otherwise switch all chiral tags
        for idx in [atom.GetIdx() for atom in self.monomer.structure.GetAtoms()]:
            self.monomer.structure.GetAtomWithIdx(idx).SetChiralTag(
                opposite_chirality(
                    self.monomer.structure.GetAtomWithIdx(idx).GetChiralTag()
                )
            )
            """if self.monomer.structure.GetAtomWithIdx(idx).GetChiralTag() == ChiralType.CHI_TETRAHEDRAL_CCW:
                self.monomer.structure.GetAtomWithIdx(idx).SetChiralTag(ChiralType.CHI_TETRAHEDRAL_CW)
            elif self.monomer.structure.GetAtomWithIdx(idx).GetChiralTag() == ChiralType.CHI_TETRAHEDRAL_CW:
                self.monomer.structure.GetAtomWithIdx(idx).SetChiralTag(ChiralType.CHI_TETRAHEDRAL_CCW)"""

    def parse_poly_carbon(self, name):
        """
        Parse long pure-carbon chains.

        Args:
            name (str): name of the long carbon chain with all the specifications of double bonds

        Returns:
            String for functional group based on the given name
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glyles', 'glyles.glycans', 'glyles.glycans.factory', 'glyles.grammar']

package_data = \
{'': ['*']}

install_requires = \
['antlr4-python3-runtime==4.9.3',
 'networkx>=2.6.3,<3.0.0',
 'numpy>=1.21.4,<2.0.0',
 'pydot>=1.4.2,<2.0.0',
 'rdkit-pypi>=2021.9.2,<2022.0.0']

setup_kwargs = {
    'name': 'glyles',
    'version': '0.4.1',
    'description': 'A tool to convert IUPAC representation of glycans into SMILES strings',
    'long_description': '# GlyLES\n\n![testing](https://github.com/kalininalab/glyles/actions/workflows/test.yaml/badge.svg) ![piwheels](https://img.shields.io/piwheels/v/glyles) ![PyPI - Downloads](https://img.shields.io/pypi/dm/glyles) [![codecov](https://codecov.io/gh/kalininalab/GlyLES/branch/main/graph/badge.svg)](https://codecov.io/gh/kalininalab/glyles)\n\nA tool to convert IUPAC representation of Glycans into SMILES representation. This repo is still in the development \nphase; so, feel free to report any errors or issues.\n\n## Specification and (current) Limitations\n\nThe exact specification we\'re referring to when talking about "IUPAC representations of glycan", is given in the \n"Notes" section of this [website](https://www.ncbi.nlm.nih.gov/glycans/snfg.html). But as this package is still in the \ndevelopment phase, not everything of the specification is implemented yet (especially not all side chains you can \nattach to monomers).\n\nThis implementation currently only works for glycans that fulfill certain properties:\n\n* The structure of the glycan can be represented as a tree of the monosaccharides with maximal branching factor 3.\n* Only basic monomers (e.g. ``Glc``, but not ``GlcNAc``) from this [website](https://www.ncbi.nlm.nih.gov/glycans/snfg.html) \n  (GalNAc is seen as modification of galactose). In addition to this list, Erythrose is parsable.\n* Some modifications can be added to the monomers, please see the [README](glyles/grammar/README.md) in the grammar\nfolder for more information on this. <br>E.g. ``GlcNAc`` is seen to be a modified version of ``Glc``. \n\n## Installation\n\nSo far, this package can only be downloaded from the python package index. So the installation with `pip` is very easy.\nJust type\n\n``````shell\npip install glyles\n``````\n\nand you\'re ready to use it as described below. Use \n\n``````shell\npip install --upgrade glyles\n``````\n\nto upgrade the glyles package to the most recent version.\n\n## Usage\n\nConvert the IUPAC into a SMILES representation using the handy `convert` method\n\n``````python\nfrom glyles.converter import convert\n\nconvert(glycan="Man(a1-2)Man", output_file="./test.txt")\n``````\n\nYou can also use the `convert_generator` method to get a generator for all SMILES:\n\n``````python\nfrom glyles.converter import convert_generator\n\nfor smiles in convert_generator(glycan_list=["Man(a1-2)Man a", "Man(a1-2)Man b"]):\n    print(smiles)\n``````\n\nIn general, the `convert` and `convert_generator` methods support the same types of input. The samples are shown\nfor `convert` but it\'s the same for `convert_generator`.\n\n* single glycan, e.g. `convert(glycan="Man(a1-2)Man)"`,\n* a list of glycans, e.g. `convert(glycan_list=["Man(a1-2)Man a", "Man(a1-2)Man"])`, and\n* a file of glycans, e.g. `convert(glycan_file="./glycans.txt")`. Here its important that the file many only contain one\n  IUPAC per line.\n* for better runtime one can also provide a generator as input, e.g. `convert(glycan_generator=some_generator)`\n\nAny output consists of tuples of the form `(input_iupac, smiles)`. `convert_generator` returns those tuples in a \ngenerator. `convert`-output is a bit different:\n\n* By default, the method returns the list of tuples.\n* You can print them to `stdout` by specifying `returning=False`.\n* The tuples can also be written `output_file`, e.g. `convert(glycan="Man(a1-2)Man", output_file="./out.csv")`.\n\n## Notation of glycans\n\nThere are multiple different notations for glycans in IUPAC. So, according to the \n[SNGF specification](https://www.ncbi.nlm.nih.gov/glycans/snfg.html), `Man(a1-4)Gal`, `Mana1-4Gal`, and `Mana4Gal` \nall describe the same disaccharide. This is also covered in this package as all three notations will be parsed into the \nsame tree of monosaccharides and result in the same SMILES string.\n\n\n## Poetry\n\nTo develop this package, we use the poetry package manager (see [here](https://python-poetry.org/) for detailed\ninstruction). It has basically the same functionality as conda but supports the package management better and also \nsupports distinguishing packages into those that are needed to use the package and those that are needed in the \ndevelopment of the package. To enable others to work on this repository, we also publish the exact \nspecifications of our poetry environment.\n',
    'author': 'Roman Joeres',
    'author_email': 'roman.joeres@helmholtz-hips.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.11',
}


setup(**setup_kwargs)



import numpy as np

from glyles.glycans.utils import UnreachableError, Tree


def enumerate_carbon(monomer):
    """
    Enumerate carbon atoms starting with the one that are also part of the root monomer and then their side chains in
    order of the original carbon atoms.

    Args:
        monomer (Monomer): object of class monomer to enumerate the carbon atoms for

    Returns:
        Nothing
    """
    # first enumerate all ring carbons and the attachments at the first and last ring atom
    c_atoms = np.where((monomer.x[:, 0] == 6) & (monomer.x[:, 2] == 1) & (monomer.x[:, 3] == 1))[0]
    ring_o = np.where((monomer.x[:, 0] == 8) & (monomer.x[:, 2] == 1) & (monomer.x[:, 3] == 1))[0]
    if ring_o.size == 0:
        next_c_id = enumerate_c_atoms(monomer, c_atoms, -1)
    else:
        next_c_id = enumerate_c_atoms(monomer, c_atoms, ring_o)

    # then iterate over all those carbons and enumerate them beginning at C1
    for c_id in range(1, next_c_id):
        c_index = int(np.where(monomer.x[:, 1] == c_id)[0])
        candidates = np.where((monomer.adjacency[c_index, :] != 0) & (monomer.x[:, 1] == 0))[0]
        if candidates.size != 0:
            for candidate in list(candidates):
                if sum(monomer.adjacency[candidate, :]) > 1:
                    next_c_id = enumerate_side_chain(monomer, c_index, candidate, next_c_id)


def enumerate_side_chain(monomer, parent, atom, next_c_id):
    """
    Not enumerating ring atoms

    Args:
        monomer (Monomer): object of class monomer to enumerate the carbon atoms for
        parent (int): RDKit ID of the parent atom of the one to look at here
        atom (int): RDKit ID of the atom to look at in this recursive call
        next_c_id (int): ID to assign to the next carbon atom

    Returns:
        Number (not RDKit ID) of the highest carbon atom in the molecule
    """
    # if side chain starts with a carbon, put a number on it
    if monomer.x[atom, 0] == 6:
        monomer.x[atom, 1] = next_c_id
        next_c_id += 1

    # identify children
    candidates = np.where((monomer.adjacency[atom, :] != 0) & (monomer.x[:, 1] == 0))[0].tolist()
    if parent in candidates:
        candidates.remove(parent)

    # and iterate over them to enumerate them in order as they appear
    if len(candidates) > 0:
        for candidate in candidates:
            next_c_id = enumerate_side_chain(monomer, atom, candidate, next_c_id)

    # return the ID of the next carbon that we see
    return next_c_id


def enumerate_c_atoms(monomer, c_atoms, ringo):
    """
    Enumerate all carbon atoms starting from the first one

    Args:
        monomer (Monomer): object of class monomer to enumerate the carbon atoms for
        c_atoms (List[int]): List of all ids of C atoms in the ring
        ringo (int): id of the oxygen atom in the ring of the monomer

    Returns:
        Number (not RDKit ID) of the highest carbon atom in the molecule
    """
    if len(c_atoms) > 0:
        # create a tree of all carbon atoms directly connected to the main ring of the monomer
        c_tree = Tree()
        stack = [(-1, c_atoms[0])]
        while len(stack) != 0:
            p_id, c_id = stack[-1]
            stack = stack[:-1]
            c_tree.add_node(c_id, p_id)

            children = np.where((monomer.adjacency[c_id, :] == 1) & (monomer.x[:, 0] == 6) & (monomer.x[:, 3] == 1))[0]
            for c in children:
                if int(c) not in c_tree.nodes:
                    stack.append((c_id, int(c)))

        # find the deepest node and rehang the tree to this node
        deepest_id, _ = c_tree.deepest_node()
        c_tree = c_tree.rehang_tree(deepest_id)
        longest_c_chain = c_tree.longest_chain()
        assert len(longest_c_chain) >= len(c_atoms), "Something went wrong, AGAIN"

        # now the two C1 candidates can be found at the ends of the longest chain
        start, end = longest_c_chain[0], longest_c_chain[-1]

        # check conditions
        start_o_conn, end_o_conn = \
            np.argwhere((monomer.adjacency[start, :] == 1) & (monomer.x[:, 0] == 8) & (monomer.x[:, 2] != 1) &
                        (monomer.x[:, 3] == 1)).squeeze().size > 0, \
            np.argwhere((monomer.adjacency[end, :] == 1) & (monomer.x[:, 0] == 8) & (monomer.x[:, 2] != 1) &
                        (monomer.x[:, 3] == 1)).squeeze().size > 0

        # decide on c1
        if start_o_conn and end_o_conn:
            if not evaluate_distance(monomer, start, end, ringo):
                longest_c_chain = reversed(longest_c_chain)
        elif end_o_conn:
            longest_c_chain = reversed(longest_c_chain)
    else:
        longest_c_chain = monomer.c1_find(monomer.structure)
    longest_c_chain = list(longest_c_chain)

    # enumerate along chain
    next_c_id = 1
    for c in longest_c_chain:
        monomer.x[c, 1] = next_c_id
        next_c_id += 1

    return next_c_id


def evaluate_distance(monomer, start, end, ringo):
    """
    Try to decide on C1 based on their distance to the oxygen in the ring

    Args:
        monomer (Monomer): object of class monomer to enumerate the carbon atoms for
        start (int): id of the first candidate for C1 atom
        end (int): id of the second candidate for C1 atom
        ringo (int): index of the oxygen atom in the ring

    Returns:
        Bool indicating that the start id is the C1 atom
    """
    adj = monomer.adjacency.copy()

    # as we have an adjacency matrix, multiply it with itself until one of the fields is non-zero
    while adj[start, ringo] == 0 and adj[end, ringo] == 0:
        adj = adj @ monomer.adjacency

    # if both fields are non-zero, we cannot decide here and have to go further
    if adj[start, ringo] > 0 and adj[end, ringo] > 0:
        return equidistant(monomer, start, end)

    return adj[start, ringo] > 0


def equidistant(monomer, start, end):
    """
    Decider for C1 in case the previous splitting rules were all tied.

    Args:
        monomer (Monomer): object of class monomer to enumerate the carbon atoms for
        start (int): id of the first candidate for C1 atom
        end (int): id of the second candidate for C1 atom

    Returns:
        Bool indicating that the start id is the C1 atom
    """
    # determine first carbon atom in the ring
    c_start_candidates = np.where((monomer.adjacency[start, :] == 1) &
                                  (monomer.x[:, 0] == 6) & (monomer.x[:, 2] == 1))[0]
    c_end_candidates = np.where((monomer.adjacency[end, :] == 1) & (monomer.x[:, 0] == 6) & (monomer.x[:, 2] == 1))[0]

    if c_start_candidates.size == 1 and c_end_candidates.size == 1:
        start_ring_c = int(c_start_candidates)
        end_ring_c = int(c_end_candidates)

        # check if those ring carbons have an attached oxygen
        start_ring_c_o_candidates = np.where((monomer.adjacency[start_ring_c, :] == 1) &
                                             (monomer.x[:, 0] == 8) & (monomer.x[:, 2] != 1))[0]
        end_ring_c_o_candidates = np.where((monomer.adjacency[end_ring_c, :] == 1) &
                                           (monomer.x[:, 0] == 8) & (monomer.x[:, 2] != 1))[0]

        if start_ring_c_o_candidates.size == 1 and end_ring_c_o_candidates.size == 1:
            raise UnreachableError("C1 atom cannot be detected")
        elif start_ring_c_o_candidates.size == 1:
            return True
    return c_start_candidates.size == 1
